"""ANTLR-generated files for parsing OpenQASM 3 files.

This package sets up its import contents to be taken from the generated files
whose ANTLR version matches the installed version of the ANTLR runtime.  The
generated files should be placed in directories called ``_<major>_<minor>``,
where `major` is 4, and `minor` is the minor version of ANTLR (e.g. if ANTLR
4.10 was used, those files should be in ``_4_10``).

The ANTLR files from more than one version of ANTLR can be present at once.  This package will
dynamically load the correct version based on the installed version of the runtime.
"""

from importlib.abc import MetaPathFinder as _MetaPathFinder, Loader as _Loader
import pathlib
import sys

if sys.version_info < (3, 10):
    from importlib_metadata import version as _version
else:
    from importlib.metadata import version as _version

# The `antlr4` package is supplied by `antlr4_python3_runtime`.
_parts = [int(x) for x in _version("antlr4_python3_runtime").split(".")]
_resolved_dir = f"_{_parts[0]}_{_parts[1]}"
_antlr_dir = pathlib.Path(__file__).parent
if not (_antlr_dir / _resolved_dir).is_dir():
    _available = [path.parent.name[1:] for path in _antlr_dir.rglob("qasm3Parser.py")]
    if not _available:
        raise ImportError("No ANTLR-generated parsers found.")
    raise ImportError(
        f"Missing ANTLR-generated parser for version '{_parts[0]}.{_parts[1]}'."
        f" Available versions: {_available!r}"
    )


class ANTLRMetaPathFinder(_MetaPathFinder):
    """Redirect module/package lookups in `openqasm3.antlr` to the concrete implementations
    pre-generated by the ANTLR version that matches the installed version of the runtime."""

    def __init__(self, version_package: str):
        top_level = __package__.rsplit(".")[0]
        # Note the extra `.` in the domain because we don't want to handle ourselves.
        self._domain = f"{top_level}._antlr."
        self._versioned = f"{top_level}._antlr.{version_package}"

    def find_spec(self, fullname, path, target=None):
        from importlib.machinery import SourceFileLoader
        from importlib.util import spec_from_loader, find_spec

        if not fullname.startswith(self._domain) or fullname.startswith(self._versioned):
            return None
        newname = f"{self._versioned}.{fullname[len(self._domain):]}"
        # Get the spec and loader for the direct path to the versioned file, and rewrap them to have
        # the unversioned module name.  The modules aren't loaded (or executed) by this, but the
        # loader is configured so that when they are, their scopes all carry the unversioned name.
        return spec_from_loader(fullname, SourceFileLoader(fullname, find_spec(newname).origin))


sys.meta_path = [ANTLRMetaPathFinder(_resolved_dir)] + sys.meta_path

# ... and now the additional content of this module.

RUNTIME_VERSION = tuple(int(x) for x in _parts)
"""The runtime-detected version of the ANTLR runtime, as a tuple like ``sys.version_info``."""

# These imports are re-directed into concrete versioned ones.  Doing them
# manually here helps stop pylint complaining.
from . import qasm3Lexer, qasm3Parser, qasm3ParserVisitor, qasm3ParserListener



import openqasm3


def test_examples(example_file):
    """Loop through all example files, verify that the ast_parser can parse the file.

    The `example_file` fixture is generated in `conftest.py`.  These tests are automatically skipped
    if the examples directly cannot be found.
    """
    with open(example_file) as f:
        source = f.read()
        openqasm3.parse(source)



"""
===================================
OpenQASM 3 Python reference package
===================================

This package contains the reference abstract syntax tree (AST) for representing
OpenQASM 3 programs, tools to parse text into this AST, and tools to manipulate
the AST.

The AST itself is in the :obj:`.ast` module.  There is a reference parser in the
:obj:`.parser` module, which requires the ``[parser]`` extra to be installed.

With the ``[parser]`` extra installed, the simplest interface to the parser is
the :obj:`~parser.parse` function.
"""

__all__ = [
    "ast",
    "visitor",
    "properties",
    "spec",
    "dump",
    "dumps",
    "parser",
    "parse",
]

__version__ = "1.0.1"

from . import ast, visitor, properties, spec

from .printer import dump, dumps

# Try to initialise the 'parsing' extra components.
try:
    import antlr4
except ModuleNotFoundError:
    pass
else:
    # Any import errors in section are of interest to the user, and should be propagated.
    del antlr4
    from . import parser
    from .parser import parse



"""
==============================================================
Generating OpenQASM 3 from an AST Node (``openqasm3.printer``)
==============================================================

.. currentmodule:: openqasm3

It is often useful to go from the :mod:`AST representation <openqasm3.ast>` of an OpenQASM 3 program
back to the textual language.  The functions and classes described here will do this conversion.

Most uses should be covered by using :func:`dump` to write to an open text stream (an open file, for
example) or :func:`dumps` to produce a string.  Both of these accept :ref:`several keyword arguments
<printer-kwargs>` that control the formatting of the output.

.. autofunction:: openqasm3.dump
.. autofunction:: openqasm3.dumps


.. _printer-kwargs:

Controlling the formatting
==========================

.. currentmodule:: openqasm3.printer

The :func:`~openqasm3.dump` and :func:`~openqasm3.dumps` functions both use an internal AST-visitor
class to operate on the AST.  This class actually defines all the formatting options, and can be
used for more low-level operations, such as writing a program piece-by-piece.  This may need access
to the :ref:`printer state <printer-state>`, documented below.

.. autoclass:: Printer
    :members:
    :class-doc-from: both

For the most complete control, it is possible to subclass this printer and override only the visitor
methods that should be modified.


.. _printer-state:

Reusing the same printer
========================

.. currentmodule:: openqasm3.printer

If the :class:`Printer` is being reused to write multiple nodes to a single stream, you will also
likely need to access its internal state.  This can be done by manually creating a
:class:`PrinterState` object and passing it in the original call to :meth:`Printer.visit`.  The
state object is mutated by the visit, and will reflect the output state at the end.

.. autoclass:: PrinterState
    :members:
"""

import contextlib
import dataclasses
import io
import functools

from typing import List, Optional, Sequence

from . import ast, properties
from .visitor import QASMVisitor

__all__ = ("dump", "dumps", "Printer", "PrinterState")


def dump(node: ast.QASMNode, file: io.TextIOBase, **kwargs) -> None:
    """Write textual OpenQASM 3 code representing ``node`` to the open stream ``file``.

    It is generally expected that ``node`` will be an instance of :class:`.ast.Program`, but this
    does not need to be the case.

    For more details on the available keyword arguments, see :ref:`printer-kwargs`.
    """
    Printer(file, **kwargs).visit(node)


def dumps(node: ast.QASMNode, **kwargs) -> str:
    """Get a string representation of the OpenQASM 3 code representing ``node``.

    It is generally expected that ``node`` will be an instance of :class:`.ast.Program`, but this
    does not need to be the case.

    For more details on the available keyword arguments, see :ref:`printer-kwargs`.
    """
    out = io.StringIO()
    dump(node, out, **kwargs)
    return out.getvalue()


@dataclasses.dataclass
class PrinterState:
    """State object for the print visitor.  This is mutated during the visit."""

    current_indent: int = 0
    """The current indentation level.  This is a non-negative integer; the actual identation string
    to be used is defined by the :class:`Printer`."""
    skip_next_indent: bool = False
    """This is used to communicate between the different levels of if-else visitors when emitting
    chained ``else if`` blocks.  The chaining occurs with the next ``if`` if this is set to
    ``True``."""

    @contextlib.contextmanager
    def increase_scope(self):
        """Use as a context manager to increase the scoping level of this context inside the
        resource block."""
        self.current_indent += 1
        try:
            yield
        finally:
            self.current_indent -= 1


def _maybe_annotated(method):
    @functools.wraps(method)
    def annotated(self: "Printer", node: ast.Statement, context: PrinterState) -> None:
        for annotation in node.annotations:
            self.visit(annotation, context)
        return method(self, node, context)

    return annotated


class Printer(QASMVisitor[PrinterState]):
    """Internal AST-visitor for writing AST nodes out to a stream as valid OpenQASM 3.

    This class can be used directly to write multiple nodes to the same stream, potentially with
    some manual control fo the state between them.

    If subclassing, generally only the specialised ``visit_*`` methods need to be overridden.  These
    are derived from the base class, and use the name of the relevant :mod:`AST node <.ast>`
    verbatim after ``visit_``."""

    def __init__(
        self,
        stream: io.TextIOBase,
        *,
        indent: str = "  ",
        chain_else_if: bool = True,
        old_measurement: bool = False,
    ):
        """
        Aside from ``stream``, the arguments here are keyword arguments that are common to this
        class, :func:`~openqasm3.dump` and :func:`~openqasm3.dumps`.

        :param stream: the stream that the output will be written to.
        :type stream: io.TextIOBase

        :param indent: the string to use as a single indentation level.
        :type indent: str, optional (two spaces).

        :param chain_else_if:  If ``True`` (default), then constructs of the form::

                if (x == 0) {
                    // ...
                } else {
                    if (x == 1) {
                        // ...
                    } else {
                        // ...
                    }
                }

            will be collapsed into the equivalent but flatter::

                if (x == 0) {
                    // ...
                } else if (x == 1) {
                    // ...
                } else {
                    // ...
                }

        :type chain_else_if: bool, optional (``True``)

        :param old_measurement: If ``True``, then the OpenQASM 2-style "arrow" measurements will be
            used instead of the normal assignments.  For example, ``old_measurement=False`` (the
            default) will emit ``a = measure b;`` where ``old_measurement=True`` would emit
            ``measure b -> a;`` instead.
        :type old_measurement: bool, optional (``False``).
        """
        self.stream = stream
        self.indent = indent
        self.chain_else_if = chain_else_if
        self.old_measurement = old_measurement

    def visit(self, node: ast.QASMNode, context: Optional[PrinterState] = None) -> None:
        """Completely visit a node and all subnodes.  This is the dispatch entry point; this will
        automatically result in the correct specialised visitor getting called.

        :param node: The AST node to visit.  Usually this will be an :class:`.ast.Program`.
        :type node: .ast.QASMNode

        :param context: The state object to be used during the visit.  If not given, a default
            object will be constructed and used.
        :type context: PrinterState
        """
        if context is None:
            context = PrinterState()
        return super().visit(node, context)

    def _start_line(self, context: PrinterState) -> None:
        if context.skip_next_indent:
            context.skip_next_indent = False
            return
        self.stream.write(context.current_indent * self.indent)

    def _end_statement(self, context: PrinterState) -> None:
        self.stream.write(";\n")

    def _end_line(self, context: PrinterState) -> None:
        self.stream.write("\n")

    def _write_statement(self, line: str, context: PrinterState) -> None:
        self._start_line(context)
        self.stream.write(line)
        self._end_statement(context)

    def _visit_statement_list(
        self, nodes: List[ast.Statement], context: PrinterState, prefix: str = ""
    ) -> None:
        self.stream.write(prefix)
        self.stream.write("{")
        self._end_line(context)
        with context.increase_scope():
            for statement in nodes:
                self.visit(statement, context)
        self._start_line(context)
        self.stream.write("}")

    def _visit_sequence(
        self,
        nodes: Sequence[ast.QASMNode],
        context: PrinterState,
        *,
        start: str = "",
        end: str = "",
        separator: str,
    ) -> None:
        if start:
            self.stream.write(start)
        for node in nodes[:-1]:
            self.visit(node, context)
            self.stream.write(separator)
        if nodes:
            self.visit(nodes[-1], context)
        if end:
            self.stream.write(end)

    def visit_Program(self, node: ast.Program, context: PrinterState) -> None:
        if node.version:
            self._write_statement(f"OPENQASM {node.version}", context)
        for statement in node.statements:
            self.visit(statement, context)

    @_maybe_annotated
    def visit_CompoundStatement(self, node: ast.CompoundStatement, context: PrinterState) -> None:
        self._start_line(context)
        self._visit_statement_list(node.statements, context)
        self._end_line(context)

    @_maybe_annotated
    def visit_Include(self, node: ast.Include, context: PrinterState) -> None:
        self._write_statement(f'include "{node.filename}"', context)

    @_maybe_annotated
    def visit_ExpressionStatement(
        self, node: ast.ExpressionStatement, context: PrinterState
    ) -> None:
        self._start_line(context)
        self.visit(node.expression, context)
        self._end_statement(context)

    @_maybe_annotated
    def visit_QubitDeclaration(self, node: ast.QubitDeclaration, context: PrinterState) -> None:
        self._start_line(context)
        self.stream.write("qubit")
        if node.size is not None:
            self.stream.write("[")
            self.visit(node.size)
            self.stream.write("]")
        self.stream.write(" ")
        self.visit(node.qubit, context)
        self._end_statement(context)

    @_maybe_annotated
    def visit_QuantumGateDefinition(
        self, node: ast.QuantumGateDefinition, context: PrinterState
    ) -> None:
        self._start_line(context)
        self.stream.write("gate ")
        self.visit(node.name, context)
        if node.arguments:
            self._visit_sequence(node.arguments, context, start="(", end=")", separator=", ")
        self.stream.write(" ")
        self._visit_sequence(node.qubits, context, separator=", ")
        self._visit_statement_list(node.body, context, prefix=" ")
        self._end_line(context)

    @_maybe_annotated
    def visit_ExternDeclaration(self, node: ast.ExternDeclaration, context: PrinterState) -> None:
        self._start_line(context)
        self.stream.write("extern ")
        self.visit(node.name, context)
        self._visit_sequence(node.arguments, context, start="(", end=")", separator=", ")
        if node.return_type is not None:
            self.stream.write(" -> ")
            self.visit(node.return_type, context)
        self._end_statement(context)

    def visit_Identifier(self, node: ast.Identifier, context: PrinterState) -> None:
        self.stream.write(node.name)

    def visit_UnaryExpression(self, node: ast.UnaryExpression, context: PrinterState) -> None:
        self.stream.write(node.op.name)
        if properties.precedence(node) >= properties.precedence(node.expression):
            self.stream.write("(")
            self.visit(node.expression, context)
            self.stream.write(")")
        else:
            self.visit(node.expression, context)

    def visit_BinaryExpression(self, node: ast.BinaryExpression, context: PrinterState) -> None:
        our_precedence = properties.precedence(node)
        # All AST nodes that are built into BinaryExpression are currently left associative.
        if properties.precedence(node.lhs) < our_precedence:
            self.stream.write("(")
            self.visit(node.lhs, context)
            self.stream.write(")")
        else:
            self.visit(node.lhs, context)
        self.stream.write(f" {node.op.name} ")
        if properties.precedence(node.rhs) <= our_precedence:
            self.stream.write("(")
            self.visit(node.rhs, context)
            self.stream.write(")")
        else:
            self.visit(node.rhs, context)

    def visit_BitstringLiteral(self, node: ast.BitstringLiteral, context: PrinterState) -> None:
        value = bin(node.value)[2:]
        if len(value) < node.width:
            value = "0" * (node.width - len(value)) + value
        self.stream.write(f'"{value}"')

    def visit_IntegerLiteral(self, node: ast.IntegerLiteral, context: PrinterState) -> None:
        self.stream.write(str(node.value))

    def visit_FloatLiteral(self, node: ast.FloatLiteral, context: PrinterState) -> None:
        self.stream.write(str(node.value))

    def visit_ImaginaryLiteral(self, node: ast.ImaginaryLiteral, context: PrinterState) -> None:
        self.stream.write(str(node.value) + "im")

    def visit_BooleanLiteral(self, node: ast.BooleanLiteral, context: PrinterState) -> None:
        self.stream.write("true" if node.value else "false")

    def visit_DurationLiteral(self, node: ast.DurationLiteral, context: PrinterState) -> None:
        self.stream.write(f"{node.value}{node.unit.name}")

    def visit_ArrayLiteral(self, node: ast.ArrayLiteral, context: PrinterState) -> None:
        self._visit_sequence(node.values, context, start="{", end="}", separator=", ")

    def visit_FunctionCall(self, node: ast.FunctionCall, context: PrinterState) -> None:
        self.visit(node.name)
        self._visit_sequence(node.arguments, context, start="(", end=")", separator=", ")

    def visit_Cast(self, node: ast.Cast, context: PrinterState) -> None:
        self.visit(node.type)
        self.stream.write("(")
        self.visit(node.argument)
        self.stream.write(")")

    def visit_DiscreteSet(self, node: ast.DiscreteSet, context: PrinterState) -> None:
        self._visit_sequence(node.values, context, start="{", end="}", separator=", ")

    def visit_RangeDefinition(self, node: ast.RangeDefinition, context: PrinterState) -> None:
        if node.start is not None:
            self.visit(node.start, context)
        self.stream.write(":")
        if node.step is not None:
            self.visit(node.step, context)
            self.stream.write(":")
        if node.end is not None:
            self.visit(node.end, context)

    def visit_IndexExpression(self, node: ast.IndexExpression, context: PrinterState) -> None:
        if properties.precedence(node.collection) < properties.precedence(node):
            self.stream.write("(")
            self.visit(node.collection, context)
            self.stream.write(")")
        else:
            self.visit(node.collection, context)
        self.stream.write("[")
        if isinstance(node.index, ast.DiscreteSet):
            self.visit(node.index, context)
        else:
            self._visit_sequence(node.index, context, separator=", ")
        self.stream.write("]")

    def visit_IndexedIdentifier(self, node: ast.IndexedIdentifier, context: PrinterState) -> None:
        self.visit(node.name, context)
        for index in node.indices:
            self.stream.write("[")
            if isinstance(index, ast.DiscreteSet):
                self.visit(index, context)
            else:
                self._visit_sequence(index, context, separator=", ")
            self.stream.write("]")

    def visit_Concatenation(self, node: ast.Concatenation, context: PrinterState) -> None:
        lhs_precedence = properties.precedence(node.lhs)
        our_precedence = properties.precedence(node)
        rhs_precedence = properties.precedence(node.rhs)
        # Concatenation is fully associative, but this package parses the AST by
        # arbitrarily making it left-associative (since the design of the AST
        # forces us to make a choice).  We emit brackets to ensure that the
        # round-trip through our printer and parser do not change the AST.
        if lhs_precedence < our_precedence:
            self.stream.write("(")
            self.visit(node.lhs, context)
            self.stream.write(")")
        else:
            self.visit(node.lhs, context)
        self.stream.write(" ++ ")
        if rhs_precedence <= our_precedence:
            self.stream.write("(")
            self.visit(node.rhs, context)
            self.stream.write(")")
        else:
            self.visit(node.rhs, context)

    @_maybe_annotated
    def visit_QuantumGate(self, node: ast.QuantumGate, context: PrinterState) -> None:
        self._start_line(context)
        if node.modifiers:
            self._visit_sequence(node.modifiers, context, end=" @ ", separator=" @ ")
        self.visit(node.name, context)
        if node.arguments:
            self._visit_sequence(node.arguments, context, start="(", end=")", separator=", ")
        self.stream.write(" ")
        self._visit_sequence(node.qubits, context, separator=", ")
        self._end_statement(context)

    def visit_QuantumGateModifier(
        self, node: ast.QuantumGateModifier, context: PrinterState
    ) -> None:
        self.stream.write(node.modifier.name)
        if node.argument is not None:
            self.stream.write("(")
            self.visit(node.argument, context)
            self.stream.write(")")

    @_maybe_annotated
    def visit_QuantumPhase(self, node: ast.QuantumPhase, context: PrinterState) -> None:
        self._start_line(context)
        if node.modifiers:
            self._visit_sequence(node.modifiers, context, end=" @ ", separator=" @ ")
        self.stream.write("gphase(")
        self.visit(node.argument, context)
        self.stream.write(")")
        if node.qubits:
            self._visit_sequence(node.qubits, context, start=" ", separator=", ")
        self._end_statement(context)

    def visit_QuantumMeasurement(self, node: ast.QuantumMeasurement, context: PrinterState) -> None:
        self.stream.write("measure ")
        self.visit(node.qubit, context)

    @_maybe_annotated
    def visit_QuantumReset(self, node: ast.QuantumReset, context: PrinterState) -> None:
        self._start_line(context)
        self.stream.write("reset ")
        self.visit(node.qubits, context)
        self._end_statement(context)

    @_maybe_annotated
    def visit_QuantumBarrier(self, node: ast.QuantumBarrier, context: PrinterState) -> None:
        self._start_line(context)
        self.stream.write("barrier")
        if node.qubits:
            self.stream.write(" ")
            self._visit_sequence(node.qubits, context, separator=", ")
        self._end_statement(context)

    @_maybe_annotated
    def visit_QuantumMeasurementStatement(
        self, node: ast.QuantumMeasurementStatement, context: PrinterState
    ) -> None:
        self._start_line(context)
        if node.target is None:
            self.visit(node.measure, context)
        elif self.old_measurement:
            self.visit(node.measure, context)
            self.stream.write(" -> ")
            self.visit(node.target, context)
        else:
            self.visit(node.target, context)
            self.stream.write(" = ")
            self.visit(node.measure, context)
        self._end_statement(context)

    def visit_ClassicalArgument(self, node: ast.ClassicalArgument, context: PrinterState) -> None:
        if node.access is not None:
            self.stream.write(
                "readonly " if node.access == ast.AccessControl.readonly else "mutable "
            )
        self.visit(node.type, context)
        self.stream.write(" ")
        self.visit(node.name, context)

    def visit_ExternArgument(self, node: ast.ExternArgument, context: PrinterState) -> None:
        if node.access is not None:
            self.stream.write(
                "readonly " if node.access == ast.AccessControl.readonly else "mutable "
            )
        self.visit(node.type, context)

    @_maybe_annotated
    def visit_ClassicalDeclaration(
        self, node: ast.ClassicalDeclaration, context: PrinterState
    ) -> None:
        self._start_line(context)
        self.visit(node.type)
        self.stream.write(" ")
        self.visit(node.identifier, context)
        if node.init_expression is not None:
            self.stream.write(" = ")
            self.visit(node.init_expression)
        self._end_statement(context)

    @_maybe_annotated
    def visit_IODeclaration(self, node: ast.IODeclaration, context: PrinterState) -> None:
        self._start_line(context)
        self.stream.write(f"{node.io_identifier.name} ")
        self.visit(node.type)
        self.stream.write(" ")
        self.visit(node.identifier, context)
        self._end_statement(context)

    @_maybe_annotated
    def visit_ConstantDeclaration(
        self, node: ast.ConstantDeclaration, context: PrinterState
    ) -> None:
        self._start_line(context)
        self.stream.write("const ")
        self.visit(node.type, context)
        self.stream.write(" ")
        self.visit(node.identifier, context)
        self.stream.write(" = ")
        self.visit(node.init_expression, context)
        self._end_statement(context)

    def visit_IntType(self, node: ast.IntType, context: PrinterState) -> None:
        self.stream.write("int")
        if node.size is not None:
            self.stream.write("[")
            self.visit(node.size, context)
            self.stream.write("]")

    def visit_UintType(self, node: ast.UintType, context: PrinterState) -> None:
        self.stream.write("uint")
        if node.size is not None:
            self.stream.write("[")
            self.visit(node.size, context)
            self.stream.write("]")

    def visit_FloatType(self, node: ast.FloatType, context: PrinterState) -> None:
        self.stream.write("float")
        if node.size is not None:
            self.stream.write("[")
            self.visit(node.size, context)
            self.stream.write("]")

    def visit_ComplexType(self, node: ast.ComplexType, context: PrinterState) -> None:
        self.stream.write("complex")
        if node.base_type is not None:
            self.stream.write("[")
            self.visit(node.base_type, context)
            self.stream.write("]")

    def visit_AngleType(self, node: ast.AngleType, context: PrinterState) -> None:
        self.stream.write("angle")
        if node.size is not None:
            self.stream.write("[")
            self.visit(node.size, context)
            self.stream.write("]")

    def visit_BitType(self, node: ast.BitType, context: PrinterState) -> None:
        self.stream.write("bit")
        if node.size is not None:
            self.stream.write("[")
            self.visit(node.size, context)
            self.stream.write("]")

    def visit_BoolType(self, node: ast.BoolType, context: PrinterState) -> None:
        self.stream.write("bool")

    def visit_ArrayType(self, node: ast.ArrayType, context: PrinterState) -> None:
        self.stream.write("array[")
        self.visit(node.base_type, context)
        self._visit_sequence(node.dimensions, context, start=", ", end="]", separator=", ")

    def visit_ArrayReferenceType(self, node: ast.ArrayReferenceType, context: PrinterState) -> None:
        self.stream.write("array[")
        self.visit(node.base_type, context)
        self.stream.write(", ")
        if isinstance(node.dimensions, ast.Expression):
            self.stream.write("#dim=")
            self.visit(node.dimensions, context)
        else:
            self._visit_sequence(node.dimensions, context, separator=", ")
        self.stream.write("]")

    def visit_DurationType(self, node: ast.DurationType, context: PrinterState) -> None:
        self.stream.write("duration")

    def visit_StretchType(self, node: ast.StretchType, context: PrinterState) -> None:
        self.stream.write("stretch")

    @_maybe_annotated
    def visit_CalibrationGrammarDeclaration(
        self, node: ast.CalibrationGrammarDeclaration, context: PrinterState
    ) -> None:
        self._write_statement(f'defcalgrammar "{node.name}"', context)

    @_maybe_annotated
    def visit_CalibrationDefinition(
        self, node: ast.CalibrationDefinition, context: PrinterState
    ) -> None:
        self._start_line(context)
        self.stream.write("defcal ")
        self.visit(node.name, context)
        self._visit_sequence(node.arguments, context, start="(", end=")", separator=", ")
        self.stream.write(" ")
        self._visit_sequence(node.qubits, context, separator=", ")
        if node.return_type is not None:
            self.stream.write(" -> ")
            self.visit(node.return_type, context)
        self.stream.write(" {")
        # At this point we _should_ be deferring to something else to handle formatting the
        # calibration grammar statements, but we're neither we nor the AST are set up to do that.
        self.stream.write(node.body)
        self.stream.write("}")
        self._end_line(context)

    @_maybe_annotated
    def visit_CalibrationStatement(
        self, node: ast.CalibrationStatement, context: PrinterState
    ) -> None:
        self._start_line(context)
        self.stream.write("cal {")
        # At this point we _should_ be deferring to something else to handle formatting the
        # calibration grammar statements, but we're neither we nor the AST are set up to do that.
        self.stream.write(node.body)
        self.stream.write("}")
        self._end_line(context)

    @_maybe_annotated
    def visit_SubroutineDefinition(
        self, node: ast.SubroutineDefinition, context: PrinterState
    ) -> None:
        self._start_line(context)
        self.stream.write("def ")
        self.visit(node.name, context)
        self._visit_sequence(node.arguments, context, start="(", end=")", separator=", ")
        if node.return_type is not None:
            self.stream.write(" -> ")
            self.visit(node.return_type, context)
        self._visit_statement_list(node.body, context, prefix=" ")
        self._end_line(context)

    def visit_QuantumArgument(self, node: ast.QuantumArgument, context: PrinterState) -> None:
        self.stream.write("qubit")
        if node.size is not None:
            self.stream.write("[")
            self.visit(node.size, context)
            self.stream.write("]")
        self.stream.write(" ")
        self.visit(node.name, context)

    @_maybe_annotated
    def visit_ReturnStatement(self, node: ast.ReturnStatement, context: PrinterState) -> None:
        self._start_line(context)
        self.stream.write("return")
        if node.expression is not None:
            self.stream.write(" ")
            self.visit(node.expression)
        self._end_statement(context)

    @_maybe_annotated
    def visit_BreakStatement(self, node: ast.BreakStatement, context: PrinterState) -> None:
        self._write_statement("break", context)

    @_maybe_annotated
    def visit_ContinueStatement(self, node: ast.ContinueStatement, context: PrinterState) -> None:
        self._write_statement("continue", context)

    @_maybe_annotated
    def visit_EndStatement(self, node: ast.EndStatement, context: PrinterState) -> None:
        self._write_statement("end", context)

    @_maybe_annotated
    def visit_BranchingStatement(self, node: ast.BranchingStatement, context: PrinterState) -> None:
        self._start_line(context)
        self.stream.write("if (")
        self.visit(node.condition, context)
        self.stream.write(")")
        self._visit_statement_list(node.if_block, context, prefix=" ")
        if node.else_block:
            self.stream.write(" else ")
            # Special handling to flatten a perfectly nested structure of
            #   if {...} else { if {...} else {...} }
            # into the simpler
            #   if {...} else if {...} else {...}
            # but only if we're allowed to by our options.
            if (
                self.chain_else_if
                and len(node.else_block) == 1
                and isinstance(node.else_block[0], ast.BranchingStatement)
                and not node.annotations
            ):
                context.skip_next_indent = True
                self.visit(node.else_block[0], context)
                # Don't end the line, because the outer-most `if` block will.
            else:
                self._visit_statement_list(node.else_block, context)
                self._end_line(context)
        else:
            self._end_line(context)

    @_maybe_annotated
    def visit_WhileLoop(self, node: ast.WhileLoop, context: PrinterState) -> None:
        self._start_line(context)
        self.stream.write("while (")
        self.visit(node.while_condition, context)
        self.stream.write(")")
        self._visit_statement_list(node.block, context, prefix=" ")
        self._end_line(context)

    @_maybe_annotated
    def visit_ForInLoop(self, node: ast.ForInLoop, context: PrinterState) -> None:
        self._start_line(context)
        self.stream.write("for ")
        self.visit(node.type)
        self.stream.write(" ")
        self.visit(node.identifier, context)
        self.stream.write(" in ")
        if isinstance(node.set_declaration, ast.RangeDefinition):
            self.stream.write("[")
            self.visit(node.set_declaration, context)
            self.stream.write("]")
        else:
            self.visit(node.set_declaration, context)
        self._visit_statement_list(node.block, context, prefix=" ")
        self._end_line(context)

    @_maybe_annotated
    def visit_SwitchStatement(self, node: ast.SwitchStatement, context: PrinterState) -> None:
        self._start_line(context)
        self.stream.write("switch (")
        self.visit(node.target, context)
        self.stream.write(") {")
        self._end_line(context)
        with context.increase_scope():
            for values, block in node.cases:
                self._start_line(context)
                self.stream.write("case ")
                self._visit_sequence(values, context, separator=", ")
                self.stream.write(" {")
                self._end_line(context)
                with context.increase_scope():
                    for statement in block.statements:
                        self.visit(statement, context)
                self._start_line(context)
                self.stream.write("}")
                self._end_line(context)
            if node.default is not None:
                self._start_line(context)
                self.stream.write("default {")
                self._end_line(context)
                with context.increase_scope():
                    for statement in node.default.statements:
                        self.visit(statement, context)
                self._start_line(context)
                self.stream.write("}")
                self._end_line(context)
        self._start_line(context)
        self.stream.write("}")
        self._end_line(context)

    @_maybe_annotated
    def visit_DelayInstruction(self, node: ast.DelayInstruction, context: PrinterState) -> None:
        self._start_line(context)
        self.stream.write("delay[")
        self.visit(node.duration, context)
        self.stream.write("]")
        if node.qubits:
            self.stream.write(" ")
            self._visit_sequence(node.qubits, context, separator=", ")
        self._end_statement(context)

    @_maybe_annotated
    def visit_Box(self, node: ast.Box, context: PrinterState) -> None:
        self._start_line(context)
        self.stream.write("box")
        if node.duration is not None:
            self.stream.write("[")
            self.visit(node.duration, context)
            self.stream.write("]")
        self._visit_statement_list(node.body, context, prefix=" ")
        self._end_line(context)

    def visit_DurationOf(self, node: ast.DurationOf, context: PrinterState) -> None:
        self.stream.write("durationof(")
        if isinstance(node.target, ast.QASMNode):
            self.visit(node.target, context)
        else:
            self._visit_statement_list(node.target, context, prefix="")
        self.stream.write(")")

    def visit_SizeOf(self, node: ast.SizeOf, context: PrinterState) -> None:
        self.stream.write("sizeof(")
        self.visit(node.target, context)
        if node.index is not None:
            self.stream.write(", ")
            self.visit(node.index)
        self.stream.write(")")

    @_maybe_annotated
    def visit_AliasStatement(self, node: ast.AliasStatement, context: PrinterState) -> None:
        self._start_line(context)
        self.stream.write("let ")
        self.visit(node.target, context)
        self.stream.write(" = ")
        self.visit(node.value, context)
        self._end_statement(context)

    @_maybe_annotated
    def visit_ClassicalAssignment(
        self, node: ast.ClassicalAssignment, context: PrinterState
    ) -> None:
        self._start_line(context)
        self.visit(node.lvalue, context)
        self.stream.write(f" {node.op.name} ")
        self.visit(node.rvalue, context)
        self._end_statement(context)

    def visit_Annotation(self, node: ast.Annotation, context: PrinterState) -> None:
        self._start_line(context)
        self.stream.write("@")
        self.stream.write(node.keyword)
        if node.command is not None:
            self.stream.write(" ")
            self.stream.write(node.command)
        self._end_line(context)

    def visit_Pragma(self, node: ast.Pragma, context: PrinterState) -> None:
        self._start_line(context)
        self.stream.write("pragma ")
        self.stream.write(node.command)
        self._end_line(context)



"""
=====================================================
Supported Specification Metadata (``openqasm3.spec``)
=====================================================

.. currentmodule:: openqasm3.spec

Metadata on the specifications supported by this package.

.. autodata:: supported_versions
"""

__all__ = ["supported_versions"]

#: A list of specification versions supported by this
#: package. Each version is a :code:`str`, e.g. :code:`'3.0'`.
supported_versions = [
    "3.0",
    "3.1",
]



import dataclasses

import pytest

import openqasm3
from openqasm3 import ast


def _remove_spans(node):
    """Return a new ``QASMNode`` with all spans recursively set to ``None`` to
    reduce noise in test failure messages."""
    if isinstance(node, list):
        return [_remove_spans(item) for item in node]
    if not isinstance(node, ast.QASMNode):
        return node
    kwargs = {}
    no_init = {}
    for field in dataclasses.fields(node):
        if field.name == "span":
            continue
        target = kwargs if field.init else no_init
        target[field.name] = _remove_spans(getattr(node, field.name))
    out = type(node)(**kwargs)
    for attribute, value in no_init.items():
        setattr(out, attribute, value)
    return out


OPERATOR_PRECEDENCE = [
    ast.BinaryOperator["||"],
    ast.BinaryOperator["&&"],
    ast.BinaryOperator["|"],
    ast.BinaryOperator["^"],
    ast.BinaryOperator["&"],
    ast.BinaryOperator["<<"],
    ast.BinaryOperator["+"],
    ast.BinaryOperator["*"],
    ast.BinaryOperator["**"],
]


class TestRoundTrip:
    """All the tests in this class are testing the round-trip properties of the "parse - print -
    parse" operation.  The test cases all need to be written in the preferred output format of the
    printer itself."""

    @pytest.mark.parametrize("indent", ["", "  ", "\t"], ids=repr)
    @pytest.mark.parametrize("chain_else_if", [True, False])
    @pytest.mark.parametrize("old_measurement", [True, False])
    def test_example_files(self, parsed_example, indent, chain_else_if, old_measurement):
        """Test that the cycle 'parse - print - parse' does not affect the generated AST of the
        example files.  Printing should just be an exercise in formatting, so should not affect how
        subsequent runs parse the file.  This also functions as something of a general integration
        test, testing much of the basic surface of the language."""
        roundtrip_ast = openqasm3.parse(
            openqasm3.dumps(
                parsed_example.ast,
                indent=indent,
                chain_else_if=chain_else_if,
                old_measurement=old_measurement,
            )
        )
        assert _remove_spans(roundtrip_ast) == _remove_spans(parsed_example.ast)

    @pytest.mark.parametrize("version_statement", ["OPENQASM 3;", "OPENQASM 3.0;"])
    def test_version(self, version_statement):
        output = openqasm3.dumps(openqasm3.parse(version_statement)).strip()
        assert output == version_statement

    def test_io_declarations(self):
        input_ = """
input int a;
input float[64] a;
input complex[float[FLOAT_WIDTH]] a;
input array[float[64], 5] a;
output bit b;
output bit[SIZE] b;
output bool b;
""".strip()
        output = openqasm3.dumps(openqasm3.parse(input_)).strip()
        assert output == input_

    def test_include(self):
        input_ = 'include "stdgates.inc";'
        output = openqasm3.dumps(openqasm3.parse(input_)).strip()
        assert output == input_

    def test_qubit_declarations(self):
        input_ = """
qubit q;
qubit[5] q;
qubit[SIZE] q;
""".strip()
        output = openqasm3.dumps(openqasm3.parse(input_)).strip()
        assert output == input_

        old_input = """
qreg q;
qreg q[5];
qreg q[SIZE];
""".strip()
        old_output = openqasm3.dumps(openqasm3.parse(old_input)).strip()
        # Note we're testing that we normalise to the new form.
        assert input_ == old_output

    def test_gate_definition(self):
        input_ = """
gate my_gate q {
}
gate my_gate(param) q {
}
gate my_gate(param1, param2) q {
}
gate my_gate q1, q2 {
}
gate my_gate q {
  x q;
}
""".strip()
        output = openqasm3.dumps(openqasm3.parse(input_), indent="  ").strip()
        assert output == input_

    def test_extern_declaration(self):
        input_ = """
extern f();
extern f() -> bool;
extern f(bool);
extern f(int[32], uint[32]);
extern f(mutable array[complex[float[64]], N_ELEMENTS]) -> int[2 * INT_SIZE];
""".strip()
        output = openqasm3.dumps(openqasm3.parse(input_)).strip()
        assert output == input_

    def test_function_declaration(self):
        input_ = """
def f() {
}
def f() -> angle[32] {
  return pi;
}
def f(int[SIZE] a) {
}
def f(qubit q1, qubit[SIZE] q2) {
}
def f(readonly array[int[32], 2] a, mutable array[uint, #dim=2] b) {
}
""".strip()
        output = openqasm3.dumps(openqasm3.parse(input_), indent="  ").strip()
        assert output == input_

    def test_unary_expression(self):
        input_ = """
!a;
-a;
~(a + a);
-a ** 2;
!true;
""".strip()
        output = openqasm3.dumps(openqasm3.parse(input_)).strip()
        assert output == input_

    def test_binary_expression(self):
        input_ = """
a * b;
a / b;
1 + 2;
1 - 2;
(1 + 2) * 3;
2 ** 8;
a << 1;
a >> b;
2 < 3;
3 >= 2;
a == b;
a != b;
a & b;
a | b;
a ^ b;
a && b;
a || b;
""".strip()
        output = openqasm3.dumps(openqasm3.parse(input_)).strip()
        assert output == input_

    def test_assignment(self):
        input_ = """
a = 1;
a = 2 * b;
a = f(4);
a += 1;
a -= a * 0.5;
a *= 2.0;
a /= 1.5;
a **= 2;
a <<= 1;
a >>= 1;
a |= f(2, 3);
a &= "101001";
""".strip()
        output = openqasm3.dumps(openqasm3.parse(input_)).strip()
        assert output == input_

    def test_index_expression(self):
        input_ = """
a[0];
a[{1, 2, 3}];
a[0][0];
a[1:2][0];
a[0][1:2];
""".strip()
        output = openqasm3.dumps(openqasm3.parse(input_)).strip()
        assert output == input_

    def test_literal(self):
        input_ = """
1;
2.0;
1.0im;
true;
false;
"1010";
"01010";
-1;
1.0ms;
1.0ns;
2.0s;
""".strip()
        output = openqasm3.dumps(openqasm3.parse(input_)).strip()
        assert output == input_

    def test_declaration(self):
        input_ = """
bool x = true;
bit x;
bit[SIZE] x;
int x = 2;
int[32] x = -5;
uint x = 0;
uint[16] x;
angle x;
angle[SIZE] x;
float x = 2.0;
float[SIZE * 2] x = 4.0;
complex[float[64]] x;
complex z;
duration a = 1.0us;
stretch b;
""".strip()
        output = openqasm3.dumps(openqasm3.parse(input_)).strip()
        assert output == input_

    def test_const_declaration(self):
        input_ = """
const bool x = true;
const int x = 2;
const int[32] x = -5;
const uint x = 0;
const uint[16] x = 0;
const angle x = pi;
const angle[SIZE] x = pi / 8;
const float x = 2.0;
const float[SIZE * 2] x = 4.0;
""".strip()
        output = openqasm3.dumps(openqasm3.parse(input_)).strip()
        assert output == input_

    def test_array_initializer(self):
        input_ = """
array[int, 2] a = {1, 2};
array[float[64], 2, 2] a = {{1.0, 0.0}, {0.0, 1.0}};
array[angle[32], 2] a = {pi, pi / 8};
array[uint[16], 4, 4] a = {b, {1, 2, 3, 4}};
array[bool, 2, 2] a = b;
""".strip()
        output = openqasm3.dumps(openqasm3.parse(input_)).strip()
        assert output == input_

    def test_alias(self):
        input_ = """
let q = a ++ b;
let q = a[1:2];
let q = a[{0, 2, 3}] ++ a[1:1] ++ a[{4, 5}];
let q = a;
""".strip()
        output = openqasm3.dumps(openqasm3.parse(input_)).strip()
        assert output == input_

    def test_function_call(self):
        input_ = """
f(1, 2, 3);
f();
f(a, b + c, a * b / c);
f(f(a));
""".strip()
        output = openqasm3.dumps(openqasm3.parse(input_)).strip()
        assert output == input_

    def test_gate_call(self):
        input_ = """
h q;
h q[0];
gphase(pi);
U(1, 2, 3) q;
U(1, 2, 3) q[0];
my_gate a, b[0:2], c;
""".strip()
        output = openqasm3.dumps(openqasm3.parse(input_)).strip()
        assert output == input_

    def test_gate_modifiers(self):
        input_ = """
ctrl @ U(1, 2, 3) a, b;
ctrl(1) @ x a, b[0];
negctrl @ U(1, 2, 3) a[0:2], b;
negctrl(2) @ h a, b, c;
pow(2) @ h a;
ctrl @ gphase(pi / 2) a, b;
inv @ h a;
inv @ ctrl @ x a, b;
ctrl(1) @ inv @ x a, b;
""".strip()
        output = openqasm3.dumps(openqasm3.parse(input_)).strip()
        assert output == input_

    def test_cast(self):
        input_ = """
int(a);
int[32](2.0);
int[SIZE](bitstring);
uint[16 + 16](a);
bit[SIZE](pi);
bool(i);
complex[float[64]](2.0);
complex[float](2.5);
float[32](1);
float(2.0);
""".strip()
        output = openqasm3.dumps(openqasm3.parse(input_)).strip()
        assert output == input_

    def test_for_loop(self):
        input_ = """
for uint i in [0:2] {
  a += 1;
}
for int[8] i in [a:b] {
  a += 1;
}
for float[64] i in [a:2 * b:c] {
  a += 1.0;
}
for uint i in {1, 2, 3} {
  a += 1;
}
for int i in {2 * j, 2 + 3 / 4, j + j} {
  a += 1;
}
for complex[float[64]] i in j {
  a += 1;
}
""".strip()
        output = openqasm3.dumps(openqasm3.parse(input_), indent="  ").strip()
        assert output == input_

    def test_while_loop(self):
        input_ = """
while (i) {
  x $0;
  i -= 1;
}
while (i == 0) {
  x $0;
  i -= 1;
}
while (!true) {
}
""".strip()
        output = openqasm3.dumps(openqasm3.parse(input_), indent="  ").strip()
        assert output == input_

    def test_if(self):
        input_ = """
if (i) {
  x $0;
}
if (true) {
  x $0;
}
if (2 + 3 == 5) {
  x $0;
}
if (!true) {
}
""".strip()
        output = openqasm3.dumps(openqasm3.parse(input_), indent="  ").strip()
        assert output == input_

    def test_else(self):
        input_ = """
if (true) {
} else {
  x $0;
}
if (true) {
} else {
  x $0;
  a = b + 2;
}
""".strip()
        output = openqasm3.dumps(openqasm3.parse(input_), indent="  ").strip()
        assert output == input_

    def test_else_if(self):
        input_ = """
if (i == 0) {
} else if (i == 1) {
} else {
  x $0;
}
if (i == 0) {
} else if (i == 1) {
  x $0;
} else if (i == 2) {
} else {
  x $1;
}
""".strip()
        output = openqasm3.dumps(openqasm3.parse(input_), indent="  ").strip()
        assert output == input_

    def test_switch_case(self):
        input_ = """
switch (i) {
  case 0 {
    x $0;
  }
  case 1, 2 {
  }
  default {
    z $0;
  }
}
""".strip()
        output = openqasm3.dumps(openqasm3.parse(input_), indent="  ").strip()
        assert output == input_

    def test_switch_no_default(self):
        input_ = """
switch (i + 1) {
  case 0 {
    x $0;
  }
  case 1, 2 {
  }
}
""".strip()
        output = openqasm3.dumps(openqasm3.parse(input_), indent="  ").strip()
        assert output == input_

    def test_jumps(self):
        input_ = """
while (true) {
  break;
  continue;
  end;
}
def f() {
  return;
}
def f() -> int[32] {
  return 2 + 3;
}
""".strip()
        output = openqasm3.dumps(openqasm3.parse(input_), indent="  ").strip()
        assert output == input_

    def test_measurement(self):
        input_ = """
measure q;
measure $0;
measure q[0];
measure q[1:3];
c = measure q;
c = measure $0;
c = measure q[0];
c = measure q[1:3];
def f() {
  return measure q;
}
def f() {
  return measure $0;
}
def f() {
  return measure q[0];
}
def f() {
  return measure q[1:3];
}
""".strip()
        output = openqasm3.dumps(
            openqasm3.parse(input_), indent="  ", old_measurement=False
        ).strip()
        assert output == input_

    def test_reset(self):
        input_ = """
reset q;
reset $0;
reset q[0];
reset q[1:3];
""".strip()
        output = openqasm3.dumps(openqasm3.parse(input_)).strip()
        assert output == input_

    def test_barrier(self):
        input_ = """
barrier q;
barrier $0;
barrier q[0];
barrier q[1:3];
barrier;
""".strip()
        output = openqasm3.dumps(openqasm3.parse(input_)).strip()
        assert output == input_

    def test_delay(self):
        input_ = """
delay[50.0ns] q;
delay[50.0ns] $0;
delay[50.0ns] q[0];
delay[50.0ns] q[1:3];
delay[2 * SIZE] q;
delay[2 * SIZE] $0;
delay[2 * SIZE] q[0];
delay[2 * SIZE] q[1:3];
delay[100.0ns];
""".strip()
        output = openqasm3.dumps(openqasm3.parse(input_)).strip()
        assert output == input_

    def test_box(self):
        input_ = """
box {
  x $0;
}
box[100.0ns] {
  x $0;
}
box[a + b] {
  x $0;
}
""".strip()
        output = openqasm3.dumps(openqasm3.parse(input_), indent="  ").strip()
        assert output == input_

    def test_duration_of(self):
        input_ = """
duration a = durationof({
  x $0;
  ctrl @ x $1, $2;
});
""".strip()
        output = openqasm3.dumps(openqasm3.parse(input_), indent="  ").strip()
        assert output == input_

    def test_pragma(self):
        input_ = """
pragma blah blah blah
pragma command
pragma namespace.command
pragma !%^* word
""".lstrip()  # The ending newline is important for pragmas.
        output = openqasm3.dumps(openqasm3.parse(input_))
        assert output == input_

    @pytest.fixture(
        params=[
            pytest.param("", id="none"),
            pytest.param("@command\n", id="single"),
            pytest.param("@command keyword\n", id="single keyword"),
            pytest.param("@command !£4&8 hello world\n", id="hard to tokenise"),
            pytest.param("@command1\n@command2 keyword\n", id="multiple"),
            pytest.param("@namespace.command1 keyword\n", id="dotted"),
            pytest.param("@namespace.foo.command1\n", id="multi-dotted"),
        ]
    )
    def annotations(self, request):
        return request.param

    @pytest.mark.parametrize(
        "statement",
        [
            pytest.param("let alias = q[1:3];", id="alias"),
            pytest.param("a = b;", id="assignment"),
            pytest.param("barrier q;", id="barrier"),
            pytest.param("box {\n}", id="box"),
            pytest.param('defcalgrammar "openpulse";', id="defcal"),
            pytest.param("int[8] a;", id="classical declaration"),
            pytest.param("const uint SIZE = 5;", id="const declaration"),
            pytest.param("def f() {\n}", id="subroutine definition"),
            pytest.param("delay[50.0ms] $0;", id="delay"),
            pytest.param("4 * 5 + 3;", id="expression"),
            pytest.param("extern f();", id="extern"),
            pytest.param("for uint i in [0:1] {\n}", id="for"),
            pytest.param("ctrl @ h q;", id="gate call"),
            pytest.param("gphase(0.5);", id="gphase call"),
            pytest.param("gate f q {\n}", id="gate definition"),
            pytest.param("if (true) {\n}", id="if"),
            pytest.param("if (true) {\n} else {\n  1;\n}", id="if-else"),
            pytest.param('include "stdgates.inc";', id="include"),
            pytest.param("input uint[8] i;", id="input declaration"),
            pytest.param("output uint[8] i;", id="output declaration"),
            pytest.param("measure $0;", id="measure"),
            pytest.param("a = measure b;", id="measure assign"),
            pytest.param("qubit[5] q;", id="qubit declaration"),
            pytest.param("reset $0;", id="reset"),
            pytest.param("while (true) {\n}", id="while"),
        ],
    )
    def test_annotations(self, statement, annotations):
        input_ = annotations + statement + "\n"
        output = openqasm3.dumps(openqasm3.parse(input_), indent="  ")
        assert output == input_


class TestExpression:
    """Test more specific features and properties of the printer when outputting expressions."""

    @pytest.mark.parametrize(
        "operator", [op for op in ast.BinaryOperator if op != ast.BinaryOperator["**"]]
    )
    def test_associativity_binary(self, operator):
        """Test that the associativity of binary expressions is respected in the output."""
        input_ = ast.Program(
            statements=[
                ast.ExpressionStatement(
                    ast.BinaryExpression(
                        lhs=ast.BinaryExpression(
                            lhs=ast.Identifier("a"),
                            op=operator,
                            rhs=ast.Identifier("b"),
                        ),
                        op=operator,
                        rhs=ast.Identifier("c"),
                    ),
                ),
                ast.ExpressionStatement(
                    ast.BinaryExpression(
                        lhs=ast.Identifier("a"),
                        op=operator,
                        rhs=ast.BinaryExpression(
                            lhs=ast.Identifier("b"),
                            op=operator,
                            rhs=ast.Identifier("c"),
                        ),
                    ),
                ),
            ],
        )
        expected = f"""
a {operator.name} b {operator.name} c;
a {operator.name} (b {operator.name} c);
""".strip()
        output = openqasm3.dumps(input_).strip()
        assert output == expected
        assert openqasm3.parse(output) == input_

    @pytest.mark.xfail(reason="Parser cannot handle bracketed concatenations")
    def test_associativity_concatenation(self):
        """The associativity of concatenation is not fully defined by the grammar or specification,
        but the printer assumes left-associativity for now."""
        input_ = ast.Program(
            statements=[
                ast.AliasStatement(
                    ast.Identifier("q"),
                    ast.Concatenation(
                        lhs=ast.Concatenation(
                            lhs=ast.Identifier("a"),
                            rhs=ast.Identifier("b"),
                        ),
                        rhs=ast.Identifier("c"),
                    ),
                ),
                ast.AliasStatement(
                    ast.Identifier("q"),
                    ast.Concatenation(
                        lhs=ast.Identifier("a"),
                        rhs=ast.Concatenation(
                            lhs=ast.Identifier("b"),
                            rhs=ast.Identifier("c"),
                        ),
                    ),
                ),
            ],
        )
        expected = """
let q = a ++ b ++ c;
let q = a ++ (b ++ c);
""".strip()
        output = openqasm3.dumps(input_).strip()
        assert output == expected
        assert openqasm3.parse(output) == input_

    @pytest.mark.xfail(reason="Currently power is still left-associative in the ANTLR grammar")
    def test_associativity_power(self):
        """Test that the right-associativity of the power expression is respected in the output."""
        input_ = ast.Program(
            statements=[
                ast.ExpressionStatement(
                    ast.BinaryExpression(
                        lhs=ast.BinaryExpression(
                            lhs=ast.Identifier("a"),
                            op=ast.BinaryOperator["**"],
                            rhs=ast.Identifier("b"),
                        ),
                        op=ast.BinaryOperator["**"],
                        rhs=ast.Identifier("c"),
                    ),
                ),
                ast.ExpressionStatement(
                    ast.BinaryExpression(
                        lhs=ast.Identifier("a"),
                        op=ast.BinaryOperator["**"],
                        rhs=ast.BinaryExpression(
                            lhs=ast.Identifier("b"),
                            op=ast.BinaryOperator["**"],
                            rhs=ast.Identifier("c"),
                        ),
                    ),
                ),
            ],
        )
        expected = f"""
(a ** b) ** c;
a ** b ** c;
""".strip()
        output = openqasm3.dumps(input_).strip()
        assert output == expected
        assert openqasm3.parse(output) == input_

    @pytest.mark.parametrize(
        ["lower", "higher"],
        [
            (lower, higher)
            for i, lower in enumerate(OPERATOR_PRECEDENCE[:-1])
            for higher in OPERATOR_PRECEDENCE[i + 1 :]
        ],
    )
    def test_precedence(self, lower, higher):
        input_ = ast.Program(
            statements=[
                ast.ExpressionStatement(
                    ast.BinaryExpression(
                        lhs=ast.BinaryExpression(
                            lhs=ast.Identifier("a"),
                            op=lower,
                            rhs=ast.Identifier("b"),
                        ),
                        op=higher,
                        rhs=ast.BinaryExpression(
                            lhs=ast.Identifier("c"),
                            op=lower,
                            rhs=ast.Identifier("d"),
                        ),
                    ),
                ),
            ],
        )
        expected = f"(a {lower.name} b) {higher.name} (c {lower.name} d);"
        output = openqasm3.dumps(input_).strip()
        assert output == expected
        assert openqasm3.parse(output) == input_


class TestOptions:
    """Test the various keyword arguments to the exporter have the desired effects."""

    @pytest.mark.parametrize("indent", ["", "  ", "\t", "    "], ids=repr)
    def test_indent(self, indent):
        input_ = f"""
def f(int[32] a) -> bool {{
{indent}return a == a;
}}
gate g(param) q {{
{indent}h q;
}}
for uint i in [0:2] {{
{indent}true;
}}
while (i) {{
{indent}i -= 1;
}}
if (i) {{
{indent}x $0;
}} else {{
{indent}x $1;
}}
box {{
{indent}x $0;
}}
durationof({{
{indent}x $0;
}});
""".strip()
        output = openqasm3.dumps(openqasm3.parse(input_), indent=indent).strip()
        assert output == input_

    @pytest.mark.parametrize("indent", ["", "  ", "\t", "   "], ids=repr)
    @pytest.mark.parametrize(
        ["outer_start", "outer_end", "allow_classical"],
        [
            pytest.param("gate f q {", "}", False, id="gate"),
            pytest.param("durationof({", "});", False, id="durationof"),
            pytest.param("def f() {", "}", True, id="function"),
            pytest.param("if (true) {", "}", True, id="if"),
            pytest.param("if (true) {\n} else {", "}", True, id="else"),
            pytest.param("box[1.0ms] {", "}", False, id="box"),
        ],
    )
    def test_indent_nested(self, indent, outer_start, outer_end, allow_classical):
        classicals = f"""
for uint i in [0:2] {{
{indent}true;
}}
while (i) {{
{indent}i -= 1;
}}
if (i) {{
{indent}x $0;
}} else {{
{indent}x $1;
}}
durationof({{
{indent}x $0;
}});
""".strip()
        quantums = f"""
box {{
{indent}x $0;
}}
""".strip()
        lines = quantums.splitlines()
        if allow_classical:
            lines.extend(classicals.splitlines())
        input_ = outer_start + "\n" + "\n".join(indent + line for line in lines) + "\n" + outer_end
        output = openqasm3.dumps(openqasm3.parse(input_), indent=indent).strip()
        assert output == input_

    def test_old_measurement(self):
        old_input = "measure q -> c;"
        output = openqasm3.dumps(openqasm3.parse(old_input), old_measurement=True).strip()
        assert output == old_input
        input_ = "c = measure q;"
        output = openqasm3.dumps(openqasm3.parse(input_), old_measurement=True).strip()
        assert output == old_input

    def test_chain_else_if(self):
        input_ = """
if (i == 0) {
} else if (i == 1) {
}
if (i == 0) {
} else if (i == 1) {
} else {
  x $0;
}
if (i == 0) {
} else if (i == 1) {
} else if (i == 2) {
}
if (i == 0) {
} else if (i == 1) {
} else if (i == 2) {
} else {
  if (i == 3) {
  }
  x $0;
}
if (i == 0) {
} else {
  if (i == 2) {
    x $0;
  } else {
    x $0;
  }
  x $0;
}
""".strip()
        output = openqasm3.dumps(openqasm3.parse(input_), indent="  ", chain_else_if=True).strip()
        assert output == input_

    def test_no_chain_else_if(self):
        input_ = """
if (i == 0) {
} else {
  if (i == 1) {
  }
}
if (i == 0) {
} else {
  if (i == 1) {
  } else {
    x $0;
  }
}
if (i == 0) {
} else {
  if (i == 1) {
  } else {
    if (i == 2) {
    }
  }
}
if (i == 0) {
} else {
  if (i == 1) {
  } else {
    if (i == 2) {
    } else {
      if (i == 3) {
      }
      x $0;
    }
  }
}
if (i == 0) {
} else {
  if (i == 2) {
    x $0;
  } else {
    x $0;
  }
  x $0;
}
""".strip()
        output = openqasm3.dumps(openqasm3.parse(input_), indent="  ", chain_else_if=False).strip()
        assert output == input_

    def test_chain_else_if_only_applies_to_else_if(self):
        input_ = """
if (i) {
} else {
  x $1;
}
if (i) {
} else {
  for uint j in [0:1] {
  }
}
if (i) {
} else {
  x $0;
  if (!i) {
  } else {
    x $1;
  }
}
""".strip()
        output = openqasm3.dumps(openqasm3.parse(input_), indent="  ", chain_else_if=True).strip()
        assert output == input_

    def test_annotations(self):
        input_ = """
@ann_1
int[32] i = 0;
if (i == 0) {
  @ann_2
  i += 1;
}
""".strip()
        output = openqasm3.dumps(openqasm3.parse(input_), indent="  ").strip()
        assert output == input_


def test_CompoundStatement():
    # This test can be replaced with a round trip test once something that
    # parses to a CompoundStatement is implemented (e.g. jakelishman's implementation
    # of the switch statement #492
    cmpnd_stmt = ast.CompoundStatement(
        statements=[
            ast.ExpressionStatement(ast.IntegerLiteral(value=1)),
            ast.ExpressionStatement(ast.IntegerLiteral(value=2)),
        ]
    )
    cmpnd_stmt.annotations = [ast.Annotation(keyword="test_annotation")]
    program = ast.Program(statements=[cmpnd_stmt])

    expected = """
@test_annotation
{
  1;
  2;
}
    """.strip()
    output = openqasm3.dumps(program).strip()
    assert output == expected



# Generated from qasm3Parser.g4 by ANTLR 4.8
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .qasm3Parser import qasm3Parser
else:
    from qasm3Parser import qasm3Parser

# This class defines a complete listener for a parse tree produced by qasm3Parser.
class qasm3ParserListener(ParseTreeListener):

    # Enter a parse tree produced by qasm3Parser#program.
    def enterProgram(self, ctx:qasm3Parser.ProgramContext):
        pass

    # Exit a parse tree produced by qasm3Parser#program.
    def exitProgram(self, ctx:qasm3Parser.ProgramContext):
        pass


    # Enter a parse tree produced by qasm3Parser#version.
    def enterVersion(self, ctx:qasm3Parser.VersionContext):
        pass

    # Exit a parse tree produced by qasm3Parser#version.
    def exitVersion(self, ctx:qasm3Parser.VersionContext):
        pass


    # Enter a parse tree produced by qasm3Parser#statement.
    def enterStatement(self, ctx:qasm3Parser.StatementContext):
        pass

    # Exit a parse tree produced by qasm3Parser#statement.
    def exitStatement(self, ctx:qasm3Parser.StatementContext):
        pass


    # Enter a parse tree produced by qasm3Parser#annotation.
    def enterAnnotation(self, ctx:qasm3Parser.AnnotationContext):
        pass

    # Exit a parse tree produced by qasm3Parser#annotation.
    def exitAnnotation(self, ctx:qasm3Parser.AnnotationContext):
        pass


    # Enter a parse tree produced by qasm3Parser#scope.
    def enterScope(self, ctx:qasm3Parser.ScopeContext):
        pass

    # Exit a parse tree produced by qasm3Parser#scope.
    def exitScope(self, ctx:qasm3Parser.ScopeContext):
        pass


    # Enter a parse tree produced by qasm3Parser#pragma.
    def enterPragma(self, ctx:qasm3Parser.PragmaContext):
        pass

    # Exit a parse tree produced by qasm3Parser#pragma.
    def exitPragma(self, ctx:qasm3Parser.PragmaContext):
        pass


    # Enter a parse tree produced by qasm3Parser#statementOrScope.
    def enterStatementOrScope(self, ctx:qasm3Parser.StatementOrScopeContext):
        pass

    # Exit a parse tree produced by qasm3Parser#statementOrScope.
    def exitStatementOrScope(self, ctx:qasm3Parser.StatementOrScopeContext):
        pass


    # Enter a parse tree produced by qasm3Parser#calibrationGrammarStatement.
    def enterCalibrationGrammarStatement(self, ctx:qasm3Parser.CalibrationGrammarStatementContext):
        pass

    # Exit a parse tree produced by qasm3Parser#calibrationGrammarStatement.
    def exitCalibrationGrammarStatement(self, ctx:qasm3Parser.CalibrationGrammarStatementContext):
        pass


    # Enter a parse tree produced by qasm3Parser#includeStatement.
    def enterIncludeStatement(self, ctx:qasm3Parser.IncludeStatementContext):
        pass

    # Exit a parse tree produced by qasm3Parser#includeStatement.
    def exitIncludeStatement(self, ctx:qasm3Parser.IncludeStatementContext):
        pass


    # Enter a parse tree produced by qasm3Parser#breakStatement.
    def enterBreakStatement(self, ctx:qasm3Parser.BreakStatementContext):
        pass

    # Exit a parse tree produced by qasm3Parser#breakStatement.
    def exitBreakStatement(self, ctx:qasm3Parser.BreakStatementContext):
        pass


    # Enter a parse tree produced by qasm3Parser#continueStatement.
    def enterContinueStatement(self, ctx:qasm3Parser.ContinueStatementContext):
        pass

    # Exit a parse tree produced by qasm3Parser#continueStatement.
    def exitContinueStatement(self, ctx:qasm3Parser.ContinueStatementContext):
        pass


    # Enter a parse tree produced by qasm3Parser#endStatement.
    def enterEndStatement(self, ctx:qasm3Parser.EndStatementContext):
        pass

    # Exit a parse tree produced by qasm3Parser#endStatement.
    def exitEndStatement(self, ctx:qasm3Parser.EndStatementContext):
        pass


    # Enter a parse tree produced by qasm3Parser#forStatement.
    def enterForStatement(self, ctx:qasm3Parser.ForStatementContext):
        pass

    # Exit a parse tree produced by qasm3Parser#forStatement.
    def exitForStatement(self, ctx:qasm3Parser.ForStatementContext):
        pass


    # Enter a parse tree produced by qasm3Parser#ifStatement.
    def enterIfStatement(self, ctx:qasm3Parser.IfStatementContext):
        pass

    # Exit a parse tree produced by qasm3Parser#ifStatement.
    def exitIfStatement(self, ctx:qasm3Parser.IfStatementContext):
        pass


    # Enter a parse tree produced by qasm3Parser#returnStatement.
    def enterReturnStatement(self, ctx:qasm3Parser.ReturnStatementContext):
        pass

    # Exit a parse tree produced by qasm3Parser#returnStatement.
    def exitReturnStatement(self, ctx:qasm3Parser.ReturnStatementContext):
        pass


    # Enter a parse tree produced by qasm3Parser#whileStatement.
    def enterWhileStatement(self, ctx:qasm3Parser.WhileStatementContext):
        pass

    # Exit a parse tree produced by qasm3Parser#whileStatement.
    def exitWhileStatement(self, ctx:qasm3Parser.WhileStatementContext):
        pass


    # Enter a parse tree produced by qasm3Parser#switchStatement.
    def enterSwitchStatement(self, ctx:qasm3Parser.SwitchStatementContext):
        pass

    # Exit a parse tree produced by qasm3Parser#switchStatement.
    def exitSwitchStatement(self, ctx:qasm3Parser.SwitchStatementContext):
        pass


    # Enter a parse tree produced by qasm3Parser#switchCaseItem.
    def enterSwitchCaseItem(self, ctx:qasm3Parser.SwitchCaseItemContext):
        pass

    # Exit a parse tree produced by qasm3Parser#switchCaseItem.
    def exitSwitchCaseItem(self, ctx:qasm3Parser.SwitchCaseItemContext):
        pass


    # Enter a parse tree produced by qasm3Parser#barrierStatement.
    def enterBarrierStatement(self, ctx:qasm3Parser.BarrierStatementContext):
        pass

    # Exit a parse tree produced by qasm3Parser#barrierStatement.
    def exitBarrierStatement(self, ctx:qasm3Parser.BarrierStatementContext):
        pass


    # Enter a parse tree produced by qasm3Parser#boxStatement.
    def enterBoxStatement(self, ctx:qasm3Parser.BoxStatementContext):
        pass

    # Exit a parse tree produced by qasm3Parser#boxStatement.
    def exitBoxStatement(self, ctx:qasm3Parser.BoxStatementContext):
        pass


    # Enter a parse tree produced by qasm3Parser#delayStatement.
    def enterDelayStatement(self, ctx:qasm3Parser.DelayStatementContext):
        pass

    # Exit a parse tree produced by qasm3Parser#delayStatement.
    def exitDelayStatement(self, ctx:qasm3Parser.DelayStatementContext):
        pass


    # Enter a parse tree produced by qasm3Parser#gateCallStatement.
    def enterGateCallStatement(self, ctx:qasm3Parser.GateCallStatementContext):
        pass

    # Exit a parse tree produced by qasm3Parser#gateCallStatement.
    def exitGateCallStatement(self, ctx:qasm3Parser.GateCallStatementContext):
        pass


    # Enter a parse tree produced by qasm3Parser#measureArrowAssignmentStatement.
    def enterMeasureArrowAssignmentStatement(self, ctx:qasm3Parser.MeasureArrowAssignmentStatementContext):
        pass

    # Exit a parse tree produced by qasm3Parser#measureArrowAssignmentStatement.
    def exitMeasureArrowAssignmentStatement(self, ctx:qasm3Parser.MeasureArrowAssignmentStatementContext):
        pass


    # Enter a parse tree produced by qasm3Parser#resetStatement.
    def enterResetStatement(self, ctx:qasm3Parser.ResetStatementContext):
        pass

    # Exit a parse tree produced by qasm3Parser#resetStatement.
    def exitResetStatement(self, ctx:qasm3Parser.ResetStatementContext):
        pass


    # Enter a parse tree produced by qasm3Parser#aliasDeclarationStatement.
    def enterAliasDeclarationStatement(self, ctx:qasm3Parser.AliasDeclarationStatementContext):
        pass

    # Exit a parse tree produced by qasm3Parser#aliasDeclarationStatement.
    def exitAliasDeclarationStatement(self, ctx:qasm3Parser.AliasDeclarationStatementContext):
        pass


    # Enter a parse tree produced by qasm3Parser#classicalDeclarationStatement.
    def enterClassicalDeclarationStatement(self, ctx:qasm3Parser.ClassicalDeclarationStatementContext):
        pass

    # Exit a parse tree produced by qasm3Parser#classicalDeclarationStatement.
    def exitClassicalDeclarationStatement(self, ctx:qasm3Parser.ClassicalDeclarationStatementContext):
        pass


    # Enter a parse tree produced by qasm3Parser#constDeclarationStatement.
    def enterConstDeclarationStatement(self, ctx:qasm3Parser.ConstDeclarationStatementContext):
        pass

    # Exit a parse tree produced by qasm3Parser#constDeclarationStatement.
    def exitConstDeclarationStatement(self, ctx:qasm3Parser.ConstDeclarationStatementContext):
        pass


    # Enter a parse tree produced by qasm3Parser#ioDeclarationStatement.
    def enterIoDeclarationStatement(self, ctx:qasm3Parser.IoDeclarationStatementContext):
        pass

    # Exit a parse tree produced by qasm3Parser#ioDeclarationStatement.
    def exitIoDeclarationStatement(self, ctx:qasm3Parser.IoDeclarationStatementContext):
        pass


    # Enter a parse tree produced by qasm3Parser#oldStyleDeclarationStatement.
    def enterOldStyleDeclarationStatement(self, ctx:qasm3Parser.OldStyleDeclarationStatementContext):
        pass

    # Exit a parse tree produced by qasm3Parser#oldStyleDeclarationStatement.
    def exitOldStyleDeclarationStatement(self, ctx:qasm3Parser.OldStyleDeclarationStatementContext):
        pass


    # Enter a parse tree produced by qasm3Parser#quantumDeclarationStatement.
    def enterQuantumDeclarationStatement(self, ctx:qasm3Parser.QuantumDeclarationStatementContext):
        pass

    # Exit a parse tree produced by qasm3Parser#quantumDeclarationStatement.
    def exitQuantumDeclarationStatement(self, ctx:qasm3Parser.QuantumDeclarationStatementContext):
        pass


    # Enter a parse tree produced by qasm3Parser#defStatement.
    def enterDefStatement(self, ctx:qasm3Parser.DefStatementContext):
        pass

    # Exit a parse tree produced by qasm3Parser#defStatement.
    def exitDefStatement(self, ctx:qasm3Parser.DefStatementContext):
        pass


    # Enter a parse tree produced by qasm3Parser#externStatement.
    def enterExternStatement(self, ctx:qasm3Parser.ExternStatementContext):
        pass

    # Exit a parse tree produced by qasm3Parser#externStatement.
    def exitExternStatement(self, ctx:qasm3Parser.ExternStatementContext):
        pass


    # Enter a parse tree produced by qasm3Parser#gateStatement.
    def enterGateStatement(self, ctx:qasm3Parser.GateStatementContext):
        pass

    # Exit a parse tree produced by qasm3Parser#gateStatement.
    def exitGateStatement(self, ctx:qasm3Parser.GateStatementContext):
        pass


    # Enter a parse tree produced by qasm3Parser#assignmentStatement.
    def enterAssignmentStatement(self, ctx:qasm3Parser.AssignmentStatementContext):
        pass

    # Exit a parse tree produced by qasm3Parser#assignmentStatement.
    def exitAssignmentStatement(self, ctx:qasm3Parser.AssignmentStatementContext):
        pass


    # Enter a parse tree produced by qasm3Parser#expressionStatement.
    def enterExpressionStatement(self, ctx:qasm3Parser.ExpressionStatementContext):
        pass

    # Exit a parse tree produced by qasm3Parser#expressionStatement.
    def exitExpressionStatement(self, ctx:qasm3Parser.ExpressionStatementContext):
        pass


    # Enter a parse tree produced by qasm3Parser#calStatement.
    def enterCalStatement(self, ctx:qasm3Parser.CalStatementContext):
        pass

    # Exit a parse tree produced by qasm3Parser#calStatement.
    def exitCalStatement(self, ctx:qasm3Parser.CalStatementContext):
        pass


    # Enter a parse tree produced by qasm3Parser#defcalStatement.
    def enterDefcalStatement(self, ctx:qasm3Parser.DefcalStatementContext):
        pass

    # Exit a parse tree produced by qasm3Parser#defcalStatement.
    def exitDefcalStatement(self, ctx:qasm3Parser.DefcalStatementContext):
        pass


    # Enter a parse tree produced by qasm3Parser#bitwiseXorExpression.
    def enterBitwiseXorExpression(self, ctx:qasm3Parser.BitwiseXorExpressionContext):
        pass

    # Exit a parse tree produced by qasm3Parser#bitwiseXorExpression.
    def exitBitwiseXorExpression(self, ctx:qasm3Parser.BitwiseXorExpressionContext):
        pass


    # Enter a parse tree produced by qasm3Parser#additiveExpression.
    def enterAdditiveExpression(self, ctx:qasm3Parser.AdditiveExpressionContext):
        pass

    # Exit a parse tree produced by qasm3Parser#additiveExpression.
    def exitAdditiveExpression(self, ctx:qasm3Parser.AdditiveExpressionContext):
        pass


    # Enter a parse tree produced by qasm3Parser#durationofExpression.
    def enterDurationofExpression(self, ctx:qasm3Parser.DurationofExpressionContext):
        pass

    # Exit a parse tree produced by qasm3Parser#durationofExpression.
    def exitDurationofExpression(self, ctx:qasm3Parser.DurationofExpressionContext):
        pass


    # Enter a parse tree produced by qasm3Parser#parenthesisExpression.
    def enterParenthesisExpression(self, ctx:qasm3Parser.ParenthesisExpressionContext):
        pass

    # Exit a parse tree produced by qasm3Parser#parenthesisExpression.
    def exitParenthesisExpression(self, ctx:qasm3Parser.ParenthesisExpressionContext):
        pass


    # Enter a parse tree produced by qasm3Parser#comparisonExpression.
    def enterComparisonExpression(self, ctx:qasm3Parser.ComparisonExpressionContext):
        pass

    # Exit a parse tree produced by qasm3Parser#comparisonExpression.
    def exitComparisonExpression(self, ctx:qasm3Parser.ComparisonExpressionContext):
        pass


    # Enter a parse tree produced by qasm3Parser#multiplicativeExpression.
    def enterMultiplicativeExpression(self, ctx:qasm3Parser.MultiplicativeExpressionContext):
        pass

    # Exit a parse tree produced by qasm3Parser#multiplicativeExpression.
    def exitMultiplicativeExpression(self, ctx:qasm3Parser.MultiplicativeExpressionContext):
        pass


    # Enter a parse tree produced by qasm3Parser#logicalOrExpression.
    def enterLogicalOrExpression(self, ctx:qasm3Parser.LogicalOrExpressionContext):
        pass

    # Exit a parse tree produced by qasm3Parser#logicalOrExpression.
    def exitLogicalOrExpression(self, ctx:qasm3Parser.LogicalOrExpressionContext):
        pass


    # Enter a parse tree produced by qasm3Parser#castExpression.
    def enterCastExpression(self, ctx:qasm3Parser.CastExpressionContext):
        pass

    # Exit a parse tree produced by qasm3Parser#castExpression.
    def exitCastExpression(self, ctx:qasm3Parser.CastExpressionContext):
        pass


    # Enter a parse tree produced by qasm3Parser#powerExpression.
    def enterPowerExpression(self, ctx:qasm3Parser.PowerExpressionContext):
        pass

    # Exit a parse tree produced by qasm3Parser#powerExpression.
    def exitPowerExpression(self, ctx:qasm3Parser.PowerExpressionContext):
        pass


    # Enter a parse tree produced by qasm3Parser#bitwiseOrExpression.
    def enterBitwiseOrExpression(self, ctx:qasm3Parser.BitwiseOrExpressionContext):
        pass

    # Exit a parse tree produced by qasm3Parser#bitwiseOrExpression.
    def exitBitwiseOrExpression(self, ctx:qasm3Parser.BitwiseOrExpressionContext):
        pass


    # Enter a parse tree produced by qasm3Parser#callExpression.
    def enterCallExpression(self, ctx:qasm3Parser.CallExpressionContext):
        pass

    # Exit a parse tree produced by qasm3Parser#callExpression.
    def exitCallExpression(self, ctx:qasm3Parser.CallExpressionContext):
        pass


    # Enter a parse tree produced by qasm3Parser#bitshiftExpression.
    def enterBitshiftExpression(self, ctx:qasm3Parser.BitshiftExpressionContext):
        pass

    # Exit a parse tree produced by qasm3Parser#bitshiftExpression.
    def exitBitshiftExpression(self, ctx:qasm3Parser.BitshiftExpressionContext):
        pass


    # Enter a parse tree produced by qasm3Parser#bitwiseAndExpression.
    def enterBitwiseAndExpression(self, ctx:qasm3Parser.BitwiseAndExpressionContext):
        pass

    # Exit a parse tree produced by qasm3Parser#bitwiseAndExpression.
    def exitBitwiseAndExpression(self, ctx:qasm3Parser.BitwiseAndExpressionContext):
        pass


    # Enter a parse tree produced by qasm3Parser#equalityExpression.
    def enterEqualityExpression(self, ctx:qasm3Parser.EqualityExpressionContext):
        pass

    # Exit a parse tree produced by qasm3Parser#equalityExpression.
    def exitEqualityExpression(self, ctx:qasm3Parser.EqualityExpressionContext):
        pass


    # Enter a parse tree produced by qasm3Parser#logicalAndExpression.
    def enterLogicalAndExpression(self, ctx:qasm3Parser.LogicalAndExpressionContext):
        pass

    # Exit a parse tree produced by qasm3Parser#logicalAndExpression.
    def exitLogicalAndExpression(self, ctx:qasm3Parser.LogicalAndExpressionContext):
        pass


    # Enter a parse tree produced by qasm3Parser#indexExpression.
    def enterIndexExpression(self, ctx:qasm3Parser.IndexExpressionContext):
        pass

    # Exit a parse tree produced by qasm3Parser#indexExpression.
    def exitIndexExpression(self, ctx:qasm3Parser.IndexExpressionContext):
        pass


    # Enter a parse tree produced by qasm3Parser#unaryExpression.
    def enterUnaryExpression(self, ctx:qasm3Parser.UnaryExpressionContext):
        pass

    # Exit a parse tree produced by qasm3Parser#unaryExpression.
    def exitUnaryExpression(self, ctx:qasm3Parser.UnaryExpressionContext):
        pass


    # Enter a parse tree produced by qasm3Parser#literalExpression.
    def enterLiteralExpression(self, ctx:qasm3Parser.LiteralExpressionContext):
        pass

    # Exit a parse tree produced by qasm3Parser#literalExpression.
    def exitLiteralExpression(self, ctx:qasm3Parser.LiteralExpressionContext):
        pass


    # Enter a parse tree produced by qasm3Parser#aliasExpression.
    def enterAliasExpression(self, ctx:qasm3Parser.AliasExpressionContext):
        pass

    # Exit a parse tree produced by qasm3Parser#aliasExpression.
    def exitAliasExpression(self, ctx:qasm3Parser.AliasExpressionContext):
        pass


    # Enter a parse tree produced by qasm3Parser#declarationExpression.
    def enterDeclarationExpression(self, ctx:qasm3Parser.DeclarationExpressionContext):
        pass

    # Exit a parse tree produced by qasm3Parser#declarationExpression.
    def exitDeclarationExpression(self, ctx:qasm3Parser.DeclarationExpressionContext):
        pass


    # Enter a parse tree produced by qasm3Parser#measureExpression.
    def enterMeasureExpression(self, ctx:qasm3Parser.MeasureExpressionContext):
        pass

    # Exit a parse tree produced by qasm3Parser#measureExpression.
    def exitMeasureExpression(self, ctx:qasm3Parser.MeasureExpressionContext):
        pass


    # Enter a parse tree produced by qasm3Parser#rangeExpression.
    def enterRangeExpression(self, ctx:qasm3Parser.RangeExpressionContext):
        pass

    # Exit a parse tree produced by qasm3Parser#rangeExpression.
    def exitRangeExpression(self, ctx:qasm3Parser.RangeExpressionContext):
        pass


    # Enter a parse tree produced by qasm3Parser#setExpression.
    def enterSetExpression(self, ctx:qasm3Parser.SetExpressionContext):
        pass

    # Exit a parse tree produced by qasm3Parser#setExpression.
    def exitSetExpression(self, ctx:qasm3Parser.SetExpressionContext):
        pass


    # Enter a parse tree produced by qasm3Parser#arrayLiteral.
    def enterArrayLiteral(self, ctx:qasm3Parser.ArrayLiteralContext):
        pass

    # Exit a parse tree produced by qasm3Parser#arrayLiteral.
    def exitArrayLiteral(self, ctx:qasm3Parser.ArrayLiteralContext):
        pass


    # Enter a parse tree produced by qasm3Parser#indexOperator.
    def enterIndexOperator(self, ctx:qasm3Parser.IndexOperatorContext):
        pass

    # Exit a parse tree produced by qasm3Parser#indexOperator.
    def exitIndexOperator(self, ctx:qasm3Parser.IndexOperatorContext):
        pass


    # Enter a parse tree produced by qasm3Parser#indexedIdentifier.
    def enterIndexedIdentifier(self, ctx:qasm3Parser.IndexedIdentifierContext):
        pass

    # Exit a parse tree produced by qasm3Parser#indexedIdentifier.
    def exitIndexedIdentifier(self, ctx:qasm3Parser.IndexedIdentifierContext):
        pass


    # Enter a parse tree produced by qasm3Parser#returnSignature.
    def enterReturnSignature(self, ctx:qasm3Parser.ReturnSignatureContext):
        pass

    # Exit a parse tree produced by qasm3Parser#returnSignature.
    def exitReturnSignature(self, ctx:qasm3Parser.ReturnSignatureContext):
        pass


    # Enter a parse tree produced by qasm3Parser#gateModifier.
    def enterGateModifier(self, ctx:qasm3Parser.GateModifierContext):
        pass

    # Exit a parse tree produced by qasm3Parser#gateModifier.
    def exitGateModifier(self, ctx:qasm3Parser.GateModifierContext):
        pass


    # Enter a parse tree produced by qasm3Parser#scalarType.
    def enterScalarType(self, ctx:qasm3Parser.ScalarTypeContext):
        pass

    # Exit a parse tree produced by qasm3Parser#scalarType.
    def exitScalarType(self, ctx:qasm3Parser.ScalarTypeContext):
        pass


    # Enter a parse tree produced by qasm3Parser#qubitType.
    def enterQubitType(self, ctx:qasm3Parser.QubitTypeContext):
        pass

    # Exit a parse tree produced by qasm3Parser#qubitType.
    def exitQubitType(self, ctx:qasm3Parser.QubitTypeContext):
        pass


    # Enter a parse tree produced by qasm3Parser#arrayType.
    def enterArrayType(self, ctx:qasm3Parser.ArrayTypeContext):
        pass

    # Exit a parse tree produced by qasm3Parser#arrayType.
    def exitArrayType(self, ctx:qasm3Parser.ArrayTypeContext):
        pass


    # Enter a parse tree produced by qasm3Parser#arrayReferenceType.
    def enterArrayReferenceType(self, ctx:qasm3Parser.ArrayReferenceTypeContext):
        pass

    # Exit a parse tree produced by qasm3Parser#arrayReferenceType.
    def exitArrayReferenceType(self, ctx:qasm3Parser.ArrayReferenceTypeContext):
        pass


    # Enter a parse tree produced by qasm3Parser#designator.
    def enterDesignator(self, ctx:qasm3Parser.DesignatorContext):
        pass

    # Exit a parse tree produced by qasm3Parser#designator.
    def exitDesignator(self, ctx:qasm3Parser.DesignatorContext):
        pass


    # Enter a parse tree produced by qasm3Parser#defcalTarget.
    def enterDefcalTarget(self, ctx:qasm3Parser.DefcalTargetContext):
        pass

    # Exit a parse tree produced by qasm3Parser#defcalTarget.
    def exitDefcalTarget(self, ctx:qasm3Parser.DefcalTargetContext):
        pass


    # Enter a parse tree produced by qasm3Parser#defcalArgumentDefinition.
    def enterDefcalArgumentDefinition(self, ctx:qasm3Parser.DefcalArgumentDefinitionContext):
        pass

    # Exit a parse tree produced by qasm3Parser#defcalArgumentDefinition.
    def exitDefcalArgumentDefinition(self, ctx:qasm3Parser.DefcalArgumentDefinitionContext):
        pass


    # Enter a parse tree produced by qasm3Parser#defcalOperand.
    def enterDefcalOperand(self, ctx:qasm3Parser.DefcalOperandContext):
        pass

    # Exit a parse tree produced by qasm3Parser#defcalOperand.
    def exitDefcalOperand(self, ctx:qasm3Parser.DefcalOperandContext):
        pass


    # Enter a parse tree produced by qasm3Parser#gateOperand.
    def enterGateOperand(self, ctx:qasm3Parser.GateOperandContext):
        pass

    # Exit a parse tree produced by qasm3Parser#gateOperand.
    def exitGateOperand(self, ctx:qasm3Parser.GateOperandContext):
        pass


    # Enter a parse tree produced by qasm3Parser#externArgument.
    def enterExternArgument(self, ctx:qasm3Parser.ExternArgumentContext):
        pass

    # Exit a parse tree produced by qasm3Parser#externArgument.
    def exitExternArgument(self, ctx:qasm3Parser.ExternArgumentContext):
        pass


    # Enter a parse tree produced by qasm3Parser#argumentDefinition.
    def enterArgumentDefinition(self, ctx:qasm3Parser.ArgumentDefinitionContext):
        pass

    # Exit a parse tree produced by qasm3Parser#argumentDefinition.
    def exitArgumentDefinition(self, ctx:qasm3Parser.ArgumentDefinitionContext):
        pass


    # Enter a parse tree produced by qasm3Parser#argumentDefinitionList.
    def enterArgumentDefinitionList(self, ctx:qasm3Parser.ArgumentDefinitionListContext):
        pass

    # Exit a parse tree produced by qasm3Parser#argumentDefinitionList.
    def exitArgumentDefinitionList(self, ctx:qasm3Parser.ArgumentDefinitionListContext):
        pass


    # Enter a parse tree produced by qasm3Parser#defcalArgumentDefinitionList.
    def enterDefcalArgumentDefinitionList(self, ctx:qasm3Parser.DefcalArgumentDefinitionListContext):
        pass

    # Exit a parse tree produced by qasm3Parser#defcalArgumentDefinitionList.
    def exitDefcalArgumentDefinitionList(self, ctx:qasm3Parser.DefcalArgumentDefinitionListContext):
        pass


    # Enter a parse tree produced by qasm3Parser#defcalOperandList.
    def enterDefcalOperandList(self, ctx:qasm3Parser.DefcalOperandListContext):
        pass

    # Exit a parse tree produced by qasm3Parser#defcalOperandList.
    def exitDefcalOperandList(self, ctx:qasm3Parser.DefcalOperandListContext):
        pass


    # Enter a parse tree produced by qasm3Parser#expressionList.
    def enterExpressionList(self, ctx:qasm3Parser.ExpressionListContext):
        pass

    # Exit a parse tree produced by qasm3Parser#expressionList.
    def exitExpressionList(self, ctx:qasm3Parser.ExpressionListContext):
        pass


    # Enter a parse tree produced by qasm3Parser#identifierList.
    def enterIdentifierList(self, ctx:qasm3Parser.IdentifierListContext):
        pass

    # Exit a parse tree produced by qasm3Parser#identifierList.
    def exitIdentifierList(self, ctx:qasm3Parser.IdentifierListContext):
        pass


    # Enter a parse tree produced by qasm3Parser#gateOperandList.
    def enterGateOperandList(self, ctx:qasm3Parser.GateOperandListContext):
        pass

    # Exit a parse tree produced by qasm3Parser#gateOperandList.
    def exitGateOperandList(self, ctx:qasm3Parser.GateOperandListContext):
        pass


    # Enter a parse tree produced by qasm3Parser#externArgumentList.
    def enterExternArgumentList(self, ctx:qasm3Parser.ExternArgumentListContext):
        pass

    # Exit a parse tree produced by qasm3Parser#externArgumentList.
    def exitExternArgumentList(self, ctx:qasm3Parser.ExternArgumentListContext):
        pass



del qasm3Parser


from . import ast

__all__ = ["precedence"]

_PRECEDENCE_TABLE = {
    ast.Concatenation: 0,
    # ... the rest of the binary operations come very early ...
    ast.UnaryExpression: 11,
    # ... power expression ...
    # "Call"-like expressions bind very tightly.
    ast.IndexExpression: 13,
    ast.FunctionCall: 13,
    ast.Cast: 13,
    ast.DurationOf: 13,
    # Identifiers/literals are the top, since you never need to put brackets
    # around a single literal or identifier.
    ast.Identifier: 14,
    ast.BitstringLiteral: 14,
    ast.BooleanLiteral: 14,
    ast.DurationLiteral: 14,
    ast.FloatLiteral: 14,
    ast.ImaginaryLiteral: 14,
    ast.IntegerLiteral: 14,
    ast.ArrayLiteral: 14,
}

_BINARY_PRECEDENCE_TABLE = {
    ast.BinaryOperator["||"]: 1,
    ast.BinaryOperator["&&"]: 2,
    ast.BinaryOperator["|"]: 3,
    ast.BinaryOperator["^"]: 4,
    ast.BinaryOperator["&"]: 5,
    # equality
    ast.BinaryOperator["=="]: 6,
    ast.BinaryOperator["!="]: 6,
    # comparsions
    ast.BinaryOperator["<"]: 7,
    ast.BinaryOperator["<="]: 7,
    ast.BinaryOperator[">"]: 7,
    ast.BinaryOperator[">="]: 7,
    # bitshifts
    ast.BinaryOperator["<<"]: 8,
    ast.BinaryOperator[">>"]: 8,
    # additive
    ast.BinaryOperator["+"]: 9,
    ast.BinaryOperator["-"]: 9,
    # multiplicative
    ast.BinaryOperator["*"]: 10,
    ast.BinaryOperator["/"]: 10,
    ast.BinaryOperator["%"]: 10,
    # ... unary expression goes here ...
    ast.BinaryOperator["**"]: 12,
}


def precedence(node: ast.QASMNode) -> int:
    """Get the integer value of the precedence level of an expression node.

    The actual numeric value has no real meaning and is subject to change
    between different versions of the AST and versions of the language.  It is
    only intended to be used as a key for comparisons between different
    expressions.

    The number is such that if an AST node representing expression ``A``
    contains a subexpression ``B``, then on output, ``B`` needs brackets around
    it if its precedence is lower than ``A``.  If ``A`` is a left-associative
    (right-associative) binary operator, then its right-hand (left-hand)
    subexpression also needs brackets if the precendence of the two are equal.
    """
    if node.__class__ in _PRECEDENCE_TABLE:
        return _PRECEDENCE_TABLE[node.__class__]
    if isinstance(node, ast.BinaryExpression):
        return _BINARY_PRECEDENCE_TABLE[node.op]
    raise ValueError(f"Node '{node}' has no expression precedence. Is it an Expression?")



import re


from openqasm3.spec import supported_versions


class TestSupportedVersions:
    SPEC_VERSION_RE = r"^(?P<major>[0-9]+)\.(?P<minor>[0-9]+)$"

    def test_supported_versions(self):
        assert supported_versions == ["3.0", "3.1"]

    def test_types(self):
        assert all(type(x) is str for x in supported_versions)  # noqa

    def test_version_formats(self):
        assert all(re.match(self.SPEC_VERSION_RE, x) for x in supported_versions)



"""
=====================================================
AST Visitors and Transformers (``openqasm3.visitor``)
=====================================================

Implementation of an example AST visitor :obj:`~QASMVisitor`, which can be
inherited from to make generic visitors of the reference AST.  Deriving from
this is :obj:`~QASMTransformer`, which is an example of how the AST can be
manipulated.
"""

from typing import Optional, TypeVar, Generic

from .ast import QASMNode

__all__ = [
    "QASMVisitor",
    "QASMTransformer",
]

T = TypeVar("T")


class QASMVisitor(Generic[T]):
    """
    A node visitor base class that walks the abstract syntax tree and calls a
    visitor function for every node found.  This function may return a value
    which is forwarded by the `visit` method.

    Modified from the implementation in ast.py in the Python standard library.
    We added the context argument to the visit method. It allows the visitor
    to hold temporary state while visiting the nodes.

    The optional context argument in visit/generic_visit methods can be used to hold temporary
    information that we do not want to hold in either the AST or the visitor themselves.
    """

    def visit(self, node: QASMNode, context: Optional[T] = None):
        """Visit a node."""
        method = "visit_" + node.__class__.__name__
        visitor = getattr(self, method, self.generic_visit)
        # The visitor method may not have the context argument.
        if context:
            return visitor(node, context)
        else:
            return visitor(node)

    def generic_visit(self, node: QASMNode, context: Optional[T] = None):
        """Called if no explicit visitor function exists for a node."""
        for value in node.__dict__.values():
            if not isinstance(value, list):
                value = [value]
            for item in value:
                if isinstance(item, QASMNode):
                    if context:
                        self.visit(item, context)
                    else:
                        self.visit(item)


class QASMTransformer(QASMVisitor[T]):
    """
    A :class:`QASMVisitor` subclass that walks the abstract syntax tree and
    allows modification of nodes.

    Modified from the implementation in ast.py in the Python standard library
    """

    def generic_visit(self, node: QASMNode, context: Optional[T] = None) -> QASMNode:
        for field, old_value in node.__dict__.items():
            if isinstance(old_value, list):
                new_values = []
                for value in old_value:
                    if isinstance(value, QASMNode):
                        value = self.visit(value, context) if context else self.visit(value)
                        if value is None:
                            continue
                        elif not isinstance(value, QASMNode):
                            new_values.extend(value)
                            continue
                    new_values.append(value)
                old_value[:] = new_values
            elif isinstance(old_value, QASMNode):
                new_node = self.visit(old_value, context) if context else self.visit(old_value)
                if new_node is None:
                    delattr(node, field)
                else:
                    setattr(node, field, new_node)
        return node
